
import tempfile
import shutil
import pydicom
import os
import warnings
import uuid


from simple_orthanc.dicom_tags import INSTANCE_LEVEL, SERIES_LEVEL, STUDY_LEVEL, PATIENT_LEVEL, OrthancDicomTag, dict_to_pydicom, safe_parse_dicom_value
from simple_orthanc import OrthancAPI, OrthancInterface
from simple_orthanc.logger import Logger
from simple_orthanc.helpers import show_progress
from simple_orthanc.constants import LOG_LEVEL, PRIVATE, ASCENDING, DESCENDING




class Orthanc(Logger):
    """
    Pythonic interface to communicate with an Orthanc Server. Selections can
    be made using dicom queries. Selection can be downloaded, deleted or read
    to numpy array.
    """
    _LOG_LEVEL = LOG_LEVEL
    warning_count = 1000
    _selection = None
    __tag_names = None

    def __init__(self, host='127.0.0.1', port=8042,
                 username=None, password=None, selection=None,
                 ipaddress=None):
        """
        Pythonic interface for an Orthanc PACS (https://www.orthanc-server.com)

        Parameters
        ----------
        host : String, optional
            host name of the orthanc server. The default is '127.0.0.1'.
        port : int, optional
            Port number of the orthanc server. The default is 8042.
        username : String, optional
            Username to connect to the orthanc server. The default is None.
        password : String, optional
            Password to connect to the orthanc server. The default is None.
        selection : TYPE, optional
            Apply a selection to the orthanc database, see the selection method
            for details. The default is None.


        Returns
        -------
        Orthanc object

        """
        Logger.__init__(self)


        if ipaddress is not None:
            warnings.warn('Use host to set set ipadress', DeprecationWarning)


        self.api = OrthancAPI(host=host,
                              port=port,
                              username=username,
                              password=password)
        

        
        self.interface = OrthancInterface(api=self.api)

        self._clear_temp()

    def __len__(self):
        return self.series_count

    def __dir__(self):
        # add dicom tag names
        
        return self._tag_names + super().__dir__()
    
    def __getattr__(self, attr):
        # return a dicom tag as attribute
        if attr in OrthancDicomTag.PYDICOM_TAGNAMES or attr.startswith(PRIVATE):
            values = self.get_dicom_tag(OrthancDicomTag(attr))
            return values
        raise AttributeError("%r object has no attribute %r" %
                              (self.__class__.__name__, attr))
        
    @property
    def _tag_names(self):
       
        if self.__tag_names is not None:
            return self.__tag_names
        
        tags = self.interface.main_dicom_tags.tolist()
        
        
        if self.patient_count == 1:
             tags += self.get_all_dicom_tags()
             
        tags = [str(tag) for tag in tags]
         
        self.__tag_names = sorted(list(set(tags)))

        return self.__tag_names

    def get_all_dicom_tags(self, include_private=True):
        """
        Returns all dicom attributes that are available for 
        the current selection as a list. 
        
        When a large number of instances are selected this may take some time.

        Returns
        -------
        TYPE
            list of strings.

        """
        
        tags = []
        for oid in self.interface.selected.Series:
           data = self.api.json_for_oid(oid=oid, level='Series')
           instance_oid = data['Instances'][0]
           tag_request = self.api.request_available_tags(instance_oid)
           available_tags = self.api.get_json(tag_request)
           
           tags += [OrthancDicomTag.from_orthanc_tag(tag)\
                    for tag in available_tags if tag not in tags]
               
        tags = [tag for tag in tags\
               if '' != str(tag) and ' ' not in str(tag)]
           
        if not include_private:
           tags = [tag for tag in tags if not str(tag).startswith(PRIVATE)]
        return tags
    
    
    def sort_by(self, tag, order=ASCENDING):
        """
         Sort the instances in the current selection based on the value of 
         the specified dicom tag. Order can be 'ascending' or 'descending'

        After sorting each successive property that is accessed will be in the
        sorted order. e.g. get_headers(), the order of the file list that is
        returned after download. 

        Parameters
        ----------
        **kwargs : dict
            dicom tag names and values that will be used to select the
            dicom images.

        Returns
        -------
        TYPE
            Orthanc Object

        """
        
        if isinstance(tag, str):
            tag = OrthancDicomTag(tag)
            
        self.raise_single_series()
            
        values = self.get_dicom_tag(tag, unique=False)
        
    
        if not isinstance(values[0], (str, float, int)):
            msg = (f'Cannot sort {tag.tagname} with values of type '
                   '{type(values[0])}')
            raise TypeError(msg)
        
        oids = self.interface.selected.Instances
        oids = [oid for _, oid in zip(values, oids)]
        
        if order == DESCENDING:
            order = oids[::-1]
        
        self.interface.selected.Instances = oids
        return self
        
    
        
    def select(self, **kwargs):
        """
         Perform a dicom query by applying a selection. Example:
            orthanc.select(PatientID='123456')
            orthanc.sekect(PatientID='123456', SeriesDescription='MyDescr')

        Parameters
        ----------
        **kwargs : dict
            dicom tag names and values that will be used to select the
            dicom images.

        Returns
        -------
        TYPE
            Orthanc Object

        """

        self.selection.update(**kwargs)

    
        self.interface.query(query=self.selection, select=True)

        self.logger.debug('Selection set %s', str(self.selection))
        
        if self.instance_count == 0:
            msg = 'No instances found for selection {0}'
            warnings.warn(msg.format(self.selection))
        
        self.__tag_names = None
        
        return self

    def refresh(self):
        """
        Refresh the local content of the orthanc server. Needed when for
        example new studies were added. SimpleOrthanc will not automatically detect
        changes on the server.

        Returns
        -------
        Orthanc.

        """
        
        self.interface.refresh()
        self.reset()
        if self.selection:
            self.select(**self.selection)
        return self

    def reset(self, tag=None):
        """
        Reset the current selection.

        Parameters
        ----------
        tag : String, optional
            If specified only this dicom tag will be removed from the current
            selection. The default is None.

        Returns
        -------
        TYPE
            DESCRIPTION.

        """

        self.logger.debug('Reset')
        if tag:
            if tag in self.selection.keys():
                self.selection.pop(tag)
                self.interface.reset()
                self.select(**self.selection)

        else:
            self._selection = None
            self.interface.reset()

        self.__tag_names = None
        return self


    @property
    def selection(self):
        """

        Current selection of items stored on the Orthanc Server.


        Returns
        -------
        dict
            Dictionary with keys the dicom tag names and values the dicom tag
            values of the current selection. e.g. {'PatientName': 'MyPatient'}.

        """

        if self._selection is None:
            self._selection = {}
        return self._selection

    @selection.setter
    def selection(self, selection):
        if selection is None:
            self.reset()
        else:
            self.select(**selection)

    
            
    def _level_count(self, level):
        return len(getattr(self.interface.selected, str(level)))
        
    @property
    def instance_count(self):
        """

        Returns
        -------
        int
            Number of instances in the current selection.

        """
        return self._level_count(INSTANCE_LEVEL)

    @property
    def series_count(self):
        """

        Returns
        -------
        int
            Number of series in the current selection.

        """
        return self._level_count(SERIES_LEVEL)

    @property
    def study_count(self):
        """

        Returns
        -------
        int
            Number of studies in the current selection.

        """
        return self._level_count(STUDY_LEVEL)

    @property
    def patient_count(self):
        """

        Returns
        -------
        int
            Number of patients in the current selection.

        """
        return self._level_count(PATIENT_LEVEL)


    def get_dicom_tag(self, tag, unique=True):
        """

        Get all dicom tag values in current selection. It may take some time
        if tag is not a main dicom tag. For non main dicom tags all dicom
        files header have to be read by the server. Main dicom tags are
        quickly obtained form the databas. 

        Parameters
        ----------
        tag : String
            Dicom tag name.
       
        unique : TYPE, optional
            Return only  unique values. The default is True. If unique is set 
            to True, the order of the values will be the same for each dicom 
            tag.

        Raises
        ------
        IndexError
            Tag not in current selection.

        Returns
        -------
        values : list, str
            List of values for each instance in the current selection. If
            unique is True, a single unique entry is reduced to a string.

        """
        
        if isinstance(tag, str):
             tag = OrthancDicomTag(tag)
        

        if self.series_count > 1 or unique:
             values = self.interface.get_dicom_tag(tag, unique=True)
        else:       
            values = self.interface.get_dicom_tag(tag, unique=False)
            
        values = [safe_parse_dicom_value(value, tag.vr, tag.vm) for value in values]
             
        return values




# =============================================================================
#  File I/O
# =============================================================================

    def upload_folder(self, folder, test_dicom=True, recursive=False):
        """
        Upload all files in a given folder to the orthanc server.

        Parameters
        ----------
        folder : string
            Folder to upload.
        test_dicom : Bool, optional
            If true files are tested if they are dicom before upload by trying
            to read the header with pydicom. Might slow down somewhat.
            The default is True.
        recursive : Bool, optional
            Find files recursively in given folder. The default is False.

        Returns
        -------
        None.

        """
        folder = os.path.abspath(folder)

        if recursive:
            files = [os.path.join(dp, f) for dp, dn, filenames \
                     in os.walk(folder) for f in filenames]
        else:
            files = os.listdir(folder)
            files = [os.path.join(folder, file) for file in files]
            files = [file for file in files if not os.path.isdir(file)]


        self.upload_file(files, test_dicom=test_dicom)

    def upload_pydicom_dataset(self, dataset):
        """
        Upload a pydicom dataset ot the orthanc server

        Parameters
        ----------
        dataset : pydicom Dataset (header and pixel_array) or a list of
        pydicom Datasets
            

        """
        
        type_error = 'Unknown type {0}, pydicom dataset expected!'
        if isinstance(dataset, pydicom.Dataset):
            dataset =[dataset]
        elif isinstance(dataset, (list, tuple)):
            pass
        else:
            raise TypeError(type_error.format(type(dataset)))
        
        folder = tempfile.mkdtemp(suffix='orthanc')
        
        for dsi in show_progress(dataset):
            if not isinstance(dsi, pydicom.Dataset):
                raise TypeError(type_error.format(type(dsi)))

            if hasattr(dsi, 'filename') and dsi.filename not in (None, ''):
                file = dsi.filename
            else:
                file = str(uuid.uuid4()) + '.dcm'
                
            dcm_file = os.path.join(folder, file) 

            pydicom.write_file(dcm_file, dsi)
            self.upload_file(dcm_file, test_dicom=False)
            try:
                os.remove(dcm_file)
            except:
                print('Failed to remove temp file {0}'.format(dcm_file))


    def upload_file(self, file, test_dicom=True, _refresh=True):
        """
        Upload a single for or a list of files to the orthanc server.

        Parameters
        ----------
        file : string or list
            File or list of files to upload.

        test_dicom : Bool, optional
            If true files are tested if they are dicom before upload by trying
            to read the header with pydicom. Might slow down somewhat.

        Returns
        -------
        None.

        """

        if isinstance(file, (list, tuple)):           
            for fi in show_progress(file):
                self.upload_file(fi, test_dicom=test_dicom, _refresh=False)
            if _refresh:
                self.refresh()
            return

        if test_dicom:
            try:
                _ = pydicom.read_file(file, stop_before_pixels=True)
            except:
                self.logger.error('File {0} is not dicom'.format(file))


        self.api.post_file(file)
        
        if _refresh:
            self.refresh()


    def download(self, folder=None):
        """
        Download all files in current selection to a folder.

        Parameters
        ----------
        folder : string, optional
            Folder to download to from orthanc server. If no folder is
            specified, files will be downloaded to a temp folder.
            The default is None.

        Returns
        -------
        files : list
            list of files with complete path that were downloaded.

        """

        if folder is None:
            folder = self._get_new_temp_folder()
        else:
            os.makedirs(folder, exist_ok=True)
        self.logger.info('Will download files to: %s', folder)
        files = []
        
        for instance_oid in show_progress(self.interface.selected.Instances):
            stream = self.api.get_stream_for_instance_oid(instance_oid)
            file = os.path.join(folder, instance_oid) + '.dcm'
            with open(file, 'wb') as out_file:
                self.logger.debug('Downloading %s', file)
                out_file.write(stream)
            files += [file]
        self.logger.info('%s files downloaded to %s', str(len(files)), folder)

        return files

# =============================================================================
#  Obtain full metadata from server
# =============================================================================

    def get_header(self, index=None, silent=True):
        """
        Return a single dicom header. If multiple instances are selected
        specify by index which header to obtain.

        Parameters
        ----------
        index : TYPE, optional
            
        silent : TYPE, optional
            If False warnings are shown for dicom tags that could not be 
            parsed. The default is True.

        Returns
        -------
        header : pydicom Dataset

        """
        if index is None:
            
            if self.instance_count == 1:
                index = 0
            else:
                self.raise_single_series()
                
        oid = self.interface.selected.Instances[index]
        dict_header = self.interface.get_header(oid)
        header = dict_to_pydicom(dict_header, silent=silent)
        return header

    def get_headers(self, index=None, silent=True):
        """
        Return pydicom headers for all instances in current selection.

        Returns
        -------
        headers : list
            list of pydicom datasets.

        """
        self.logger.info('Will read {0} headers..'.format(self.instance_count))
        
        dict_headers = self.interface.get_selected_headers()
        pydicom_headers = [dict_to_pydicom(dct, silent=silent)\
                           for dct in dict_headers]
        return pydicom_headers


    def delete_selected_patient(self):
        """
        Delete selected patient on orthanc server

        """
        if self.patient_count != 1:
            msg = '{0} Patients in selection, can only delete single patient'
            raise IndexError(msg.format(self.patient_count))
        self.api.delete_oid(oid=self.interface.Patients[0], level=PATIENT_LEVEL)
        
        self.reset()
        self.refresh()
            

    def delete_selected_study(self):
        """
        Delete selected study on orthanc server

        """
        if self.study_count != 1:
            msg = '{0} Studies in selection, can only delete single study'
            raise IndexError(msg.format(self.study_count))

        self.api.delete_oid(oid=self.interface.Studies[0], level=STUDY_LEVEL)
        self.reset()
        self.refresh()
        
    def delete_selected_serie(self):
        """
        Delete selected patient on orthanc server

        """
        if self.series_count > 1:
            msg = '{0} Studies in selection, can only delete single study'
            raise IndexError(msg.format(self.study_count))

        self.api.delete_oid(oid=self.interface.Series[0], level=SERIES_LEVEL)
        self.reset()
        self.refresh()

    @property
    def _temp_folder(self):
        # return root temp folder
        folder = os.path.join(tempfile.gettempdir(), 'SimpleOrthanc')
        os.makedirs(folder, exist_ok=True)
        return folder

    def _clear_temp(self):
        # remove temp folder
        shutil.rmtree(self._temp_folder, ignore_errors=True)

    def _get_new_temp_folder(self):
        # create a new temporary folder to store files
        # folders are numbered from 0
        def make_int(value):
            try:
                value = int(value)
            except ValueError:
                value = 0
            return value


        folder = self._temp_folder

        n = [int(subfolder) for subfolder in os.listdir(folder)]

        if len(n) == 0:
            n = [0]

        n = max(n) + 1

        new_temp_folder =  os.path.join(folder, str(n))

        os.makedirs(new_temp_folder)

        return new_temp_folder

    
    def get_array(self, apply_suv_rescale=True):            
        """
        Get the numpy array from the dicom files in the current selection.
        A single series must be selected. When a series has mulitple slices
        (instances) all slices are read. The array has shape MxN*S*c
        where M and N are the number of pixels, S the number of slices and
        c the number of color channels (1 (GrayScale), 3 (Color))

        Parameters
        ----------
        apply_suv_rescale : Bool, optional
            Use header information to scale the array to SUV values. 
            Works only if modality is PET or NM. The default is True.

       

        Returns
        -------
        arr : numpy.ndarray

        """
        if self.series_count == 0:
            raise IndexError('No series selected!')
            
        if self.series_count > 1:
            raise IndexError('Cannot load multiple series')
        
        if self.instance_count == 1:
            arr = self.interface.get_numpy(self.interface.selected.Instances[0])
        else:
            arr = self.interface.get_numpy(self.interface.selected.Series[0])
            
        return arr
    
    
    def raise_position_information(self):
        all_tags = self.get_all_dicom_tags()
        if 'ImagePositionPatient' not in all_tags:
            raise KeyError('Dicom header does not have the ImagePositionPatient tag')
        if 'ImageOrientationPatient' not in all_tags:
            raise KeyError('Dicom header does not have the ImageOrientationPatient tag')
            
    def raise_single_series(self):
        if self.series_count > 1:
            raise IndexError('Multiple Series selected. Selection should contain a single series')
        if self.series_count == 0:
            raise IndexError('No Series selected. Selection should contain a single series')
            







import pydicom
import numpy as np
from simple_orthanc import Orthanc

try:
    import SimpleITK as sitk
    import sitktools
except ImportError:
    sitk = None
    
try:
    import simple_slice_viewer as ssv
except ImportError:
    ssv = None
    
# Functions and objects related to correct reading of dicom to a (3D) np array
# or (3D) SimpleITK image. RescaleSlope and Intercept are applied. PET SUV
# scaling when requested. For SimpleITK images the correct geometry is read
# from ImagePositionPatient and ImageOrientationPatient.

# (!) Functions are tested on a limited set of dicom images and modalities.
# Mainly PET, CT and some simple MR images (!)

IMAGEPOSITIONPATIENT = 'ImagePositionPatient'
IMAGEORIENTATIONPATIENT = 'ImageOrientationPatient'

class OrthancImage(Orthanc):
    # extension for the orthanc class to read images directly
    
    
# =============================================================================
#     Obtain image data from server
# =============================================================================
    def __init__(self, *args, **kwargs):
        if sitktools is None or sitk is None:
            raise ImportError('Need sitktools ans SimpleITK package for OrthancImage')
        
        
        super().__init__(*args, **kwargs)
        
    def get_array(self, apply_suv_rescale=True):   
        array = super().get_array()
        if apply_suv_rescale:
            array = self._apply_suv_rescale(array)
        return array
        
    def sort_slices(self):
        """
        Sort instances by slice location. Works when a single series is 
        selected. ImagePositionPatient and ImageOrientation must be in the
        dicom header. 
        
        After sorting each successive property that is accessed will be in the
        sorted order. e.g. get_headers(), the order of the file list that is
        returned after download. 

        Returns
        -------
        TYPE
            Orthanc Object

        """
        
        self.raise_position_information()
        self.raise_single_series()
            
        positions = self.get_dicom_tag(IMAGEPOSITIONPATIENT, unique=False)
        orientation = self.get_dicom_tag(IMAGEORIENTATIONPATIENT)[0]

        _, sequence = sitktools.sitk_from_array.arg_sort_positions(positions, orientation)
                                                
        oids = self.interface.selected.Instances
        sorted_oids = [oids[i] for i in sequence]
        self.interface.selected.Instances = sorted_oids
        return self

    def get_image(self, apply_suv_rescale=True):
        """
        Return sitk image for current selection.

        Raises
        ------
        IndexError
            Raised when multiple series are in current selection.

        Returns
        -------
        image : SimpleITK Image

        """

        if self.series_count != 1:
            msg = 'Can only read a single series, {0} series selected'
            raise IndexError(msg.format(self.series_count))

        if self.interface.supports_numpy:
            image = self.get_image_from_server()
        else:
            image = self.get_image_from_download()

        return image
    
    def _apply_suv_rescale(self, im_array):
        header = self.get_header(index=0)
        if header.Modality != 'PT':
            return im_array
    
        im_array *= sitktools.suv_scale_factor(header)
        
        return im_array
            
    def get_image_from_download(self, apply_suv_rescale=True):        
        files = self.download() 
        image = sitktools.read_files(files, SUV=apply_suv_rescale)
        return image
        
    def get_image_from_server(self, apply_suv_rescale=True):
        arr = self.get_array(apply_suv_rescale=apply_suv_rescale)
            
        all_tags = self.get_all_dicom_tags()
        
        for tag in ('ImagePositionPatient', 'ImageOrientationPatient',
                    'PixelSpacing'):
            if tag not in all_tags:
                raise KeyError(f'Dicom header does not have the {tag} tag')
                
        position    = self.get_dicom_tag('ImagePositionPatient', unique=False)
        orientation = self.ImageOrientationPatient[0]
        spacing     = self.PixelSpacing[0]
        
        image  = sitktools.sitk_from_array.array_to_sitk(arr, position, orientation, spacing)
        return image
    
    def display(self, apply_suv_rescale=True):
        if ssv is None:
            raise ImportError('SimpleSliceViewer not installed!')
        image = self.get_image(apply_suv_rescale=apply_suv_rescale)
        ssv.display(image)



    




import json
import requests
from requests.auth import HTTPBasicAuth
from simple_orthanc.dicom_tags import INSTANCE_LEVEL, SERIES_LEVEL, STUDY_LEVEL, PATIENT_LEVEL
from simple_orthanc.logger import Logger


class OrthancAPI(Logger):
    """
    Construct and execute http requests for a defined orthanc server.
    """
    MAIN_DICOM_TAGS = 'MainDicomTags'
    PARENT = 'Parent'
    VERSION = 'Version'
    
    _LOG_LEVEL = Logger.LEVEL_ERROR
    _version = None
    _main_dicom_tags = None
    def __init__(self, host='127.0.0.1', port=8042,
                 username=None, password=None):
        """
        Specify ipaddress, port, username(optional) and password(optional) for
        an orthanc server.
        """
        super().__init__()

        self.host = host
        self.port = port
        self.server = 'http://{host}:{port}'.format(host=host, port=port)
        self.username = username
        self.password = password
        
        r = self.get_json(self.request_system())
        self.version = r[self.VERSION]
        if self.MAIN_DICOM_TAGS in r.keys():
            self.main_dicom_tags = r[self.MAIN_DICOM_TAGS]
        else:
            self.main_dicom_tags = self._default_main_dicomtags_response()
        
       
    @property
    def headers(self):
        """
        Headers to include in http requests.
        """
        headers = {'content-type' : 'application/dicom'}
        return headers

    @property
    def auth(self):
        """
        Return basic authentication header for set username and password.
        """

        if self.username is not None:
            auth = HTTPBasicAuth(self.username, self.password)
        else:
            auth = None

        return auth

    def request_peers(self):
        return '{server}/peers?expand'.format(server=self.server)

    def request_send(self, peer=''):
        return '{server}/peers/{peer}/store'.format(server=self.server, peer=peer)

    def request_level(self, level=None):
        """
        Construct request for a specific level (Patient, Study, Series, Instance)

        example:
        OrthancAPI().request_level(level='Series')

        Subsequent execution of the request will return all oids for all series
        """
        level = str(level).lower()
        return '{server}/{level}'.format(server=self.server, level=level)

    def request_query(self):
        """
        Construct a http dicom query.

        query: Dictionary with keys valid dicom tags

        example:
        OrthancAPI().request_query(query={'PatientName':' MyPatient'})
        """
        return '{server}/tools/find'.format(server=self.server)

    def request_oid(self, oid=None, level=None):
        """
        Construct a http request to retrieve metadata for a specific oid.

        example:
        OrthancAPI().request_oid(oid=my_oid, level='Series')

        Subsequent execution of the request will return all Dicom meta data for
        the given oid on series level.
        """

        level = str(level).lower()
        return '{server}/{level}/{oid}'.format(server=self.server,
                                               level=level, oid=oid)



    def request_file(self, instance_oid):
        """
        Construct http request to retrieve the contents of a dicom file

        example:
        OrthancAPI().request_file(instance_oid)

        Subsequent execution of the request will return the binary data for
        the dicom file that matches with the instance_oid.
        """
        req = '{server}/instances/{instance_oid}/file'
        return req.format(server=self.server, instance_oid=instance_oid)

    def request_simplified_tags(self, oid):
        """
        Request to obtain main dicom tags.
        """
        req = '{server}/instances/{instance_oid}/simplified-tags'
        return req.format(server=self.server, instance_oid=oid)

    def request_raw_tag(self, oid=None, tag=None):
        """
        Request to obtain a value from any dicom tag.
        """

        request = '{server}/instances/{oid}/content/{tag}/'
        return request.format(server=self.server, oid=oid, tag=tag)

    def request_available_tags(self, instance_oid):
        """
        Request to obtain available dicom tags for a single instance.
        """
        request = '{server}/instances/{oid}/content/'
        return request.format(server=self.server, oid=instance_oid)
    
    def request_numpy(self, oid, level):
        level = level.lower()
        request = '{server}/{level}/{oid}/numpy'
        return request.format(server=self.server, oid=oid, level=level)
    
    def request_system(self):
        return '{server}/system'.format(server=self.server)
    

    def post_file(self, file):
        """
        Upload a dicom file to orthanc.
        """

        with open(file, "rb") as fid:
            content = fid.read()

        self.logger.debug("Uploading %s", file)

        request = self.request_level(level=INSTANCE_LEVEL)

        return self.post(request, content)

    def send_oids(self, request, oids):
        if not isinstance(oids, str):
            oids = json.dumps(oids)
        return self.post(request, data=oids)

    def delete(self, request):
        """
        Execute http delete request for orthanc server.
        """
        msg = 'Executing delete %s.'
        self.logger.debug(msg, request)

        # requests.get(request, headers=self.headers, auth=self.auth,
        #                     data=data)
        r = requests.delete(request, headers=self.headers, auth=self.auth)
        r.raise_for_status()


    def get(self, request, data=None):
        """
        Execute http get request for the orthanc server
        """
        self.logger.debug('Executing get %s', request)
        r = requests.get(request, headers=self.headers, auth=self.auth,
                         data=data)
        r.raise_for_status()
        return r

    def post(self, request, data):
        """
        Execute http post request with data
        """
        msg = 'Executing post %s with data %s'
        self.logger.debug(msg, request, data)
        r = requests.post(request, data=data, headers=self.headers,
                             auth=self.auth)
        r.raise_for_status()
        return r

    def get_json(self, request):
        """
        Execute get request and return response as json.
        """
        return self.get(request).json()

    def post_json(self, request, data):
        """
        Execute post request and return response as json.
        """

        return self.post(request, data).json()

    def get_stream(self, request):
        """
        Execute get request and return binary response.
        """
        return self.get(request).content

    def get_all_oids(self):
        """
        Get all oids on the Orthanc Server.
        """
        result = {}

        for level in [PATIENT_LEVEL, STUDY_LEVEL, SERIES_LEVEL, INSTANCE_LEVEL]:
            request = self.request_level(level)
            result[str(level)] = self.get_json(request)

        return result

    def delete_oid(self, oid, level=None):
        """
        Delete item by oid from Orthanc server.
        """
        request = self.request_oid(oid=oid, level=level)
        return self.delete(request)

    def get_stream_for_instance_oid(self, instance_oid):
        # request file content for a specific instance oid
        request = self.request_file(instance_oid)
        result = self.get_stream(request)
        return result

    def get_simplified_tags(self, instance_oid):
        response = self.get(self.request_simplified_tags(instance_oid))
        return response.json()

    def get_tag(self, instance_oid=None, tag=None):
        request = self.request_raw_tag(oid=instance_oid, tag=tag)
        result = self.get(request).content
        result = self.decode_tag_request(result)
        if result == '':
            if tag not in self.get_simplified_tags(instance_oid).keys():
                raise KeyError('Instance doesn''t have Tag {0}'.format(tag))
        return result

    def json_for_oid(self, level=None, oid=None):
        # return json data for a specific oid and level
        request = self.request_oid(oid=oid, level=level)

        result = self.get_json(request)
        return result
    @staticmethod
    def _default_main_dicomtags_response():
        return {'Instance': ('0008,0012;0008,0013;0008,0018;0020,0012;'
                             '0020,0013;0020,0032;0020,0037;0020,0100;'
                             '0020,4000;0028,0008;0054,1330'),
                'Patient': '0010,0010;0010,0020;0010,0030;0010,0040;0010,1000',
                'Series': ('0008,0021;0008,0031;0008,0060;0008,0070;0008,1010;'
                          '0008,103e;0008,1070;0018,0010;0018,0015;0018,0024;'
                          '0018,1030;0018,1090;0018,1400;0020,000e;0020,0011;'
                          '0020,0037;0020,0105;0020,1002;0040,0254;0054,0081;'
                          '0054,0101;0054,1000'),
                'Study': ('0008,0020;0008,0030;0008,0050;0008,0080;0008,0090;'
                          '0008,1030;0020,000d;0020,0010;0032,1032;0032,1060')}
    @staticmethod
    def decode_tag_request(result):
        if hasattr(result, 'decode'):
            result = result.decode()
        if isinstance(result, str):
            result = result.strip()
        return result



import logging

DEFAULT_LOG_LEVEL = logging.ERROR

class Logger():
    _logger = None
    _APP_NAME = 'Application'
    _LOG_FILE = None
    _LOG_LEVEL = DEFAULT_LOG_LEVEL
    _LOG_TO_CONSOLE = True
    _LOG_FORMAT = '%(asctime)s - %(name)s - %(levelname)s - %(message)s'

    LEVEL_DEBUG = logging.DEBUG
    LEVEL_INFO  = logging.INFO
    LEVEL_ERROR = logging.ERROR

    def __init__(self, log_level =  None, app_name = None):
        if log_level is not None:
            self._LOG_LEVEL = log_level
        if app_name is not None:
            self._APP_NAME = app_name
        else:
            self._APP_NAME = self.__class__.__name__
        self.logger.setLevel(self._LOG_LEVEL)
    @property
    def logger(self):
        if self._logger is None:
            self._logger = applogger(self._APP_NAME,
                                     fname = self._LOG_FILE,
                                     log_level = self._LOG_LEVEL,
                                     log_to_console = self._LOG_TO_CONSOLE,
                                     log_format = self._LOG_FORMAT)
        return self._logger


def applogger(app_name, fname = None, log_level = DEFAULT_LOG_LEVEL,
                       log_to_console = True, log_format = None):
  """ Create a simple logger object for a specific application name (app_name).
      log_level and log_to_console can be set. Log to file is done when fname
      is a valid filename """
  logger = logging.getLogger(app_name)
  logger.setLevel(log_level)
  logger.handlers = []
  # create file handler which logs even debug messages
  if log_format is None:
      log_format = '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
  formatter = logging.Formatter(log_format)

  if not(fname is None):
    fh = logging.FileHandler(fname)
    fh.setLevel(log_level)
    fh.setFormatter(formatter)
    logger.addHandler(fh)

  if log_to_console:
    ch = logging.StreamHandler()
    ch.setLevel(log_level)
    ch.setFormatter(formatter)
    logger.addHandler(ch)

  return logger



from simple_orthanc.constants import UNKNOWN
import pydicom
from dataclasses import dataclass
from pydicom import Dataset
from dateutil import parser
from datetime import datetime, timedelta
    

@dataclass
class DicomLevel:
    level: str
    _HIARCHY = ['Patients', 'Studies', 'Series', 'Instances']
    
    _PLURAL_TO_SINGULAR = {'Patients':  'Patient',
                          'Studies':   'Study',
                          'Series':    'Series',
                          'Instances': 'Instance'}
    
    _SINGULAR_TO_PLURAL = {'Patient':  'Patients',
                          'Study':   'Studies',
                         ' Serie':    'Series',
                          'Instance': 'Instances'}
    
    @property
    def levels(self):
        return [DicomLevel(level) for level in self._HIARCHY]
    
    @property
    def child_level(self):
        level_num = int(self) + 1
        if level_num >= len(self.levels):
            return None
        else:
            return self.levels[level_num]
    @property
    def parent_level(self):
        level_num = int(self) - 1
        if level_num < 0:
            return None
        else:
            return self.levels[level_num]
    @property
    def parent_levels(self):
        return [level for level in self.levels if int(level) < int(self)]    

    @property
    def child_levels(self):
        return [level for level in self.levels if int(level) > int(self)]    
        
    def _format_other(self, other):
        if isinstance(other, DicomLevel):
            pass
        elif isinstance(other, str):
            other = DicomLevel(other)
        else:
            raise TypeError()
        return other
    
    def __hash__(self):
        return str(self).__hash__()
    
    def __gt__(self, other):
        other = self._format_other(other)
        return int(self) < int(other)
    
    def __ge__(self, other):
        other = self._format_other(other)
        return int(self) <= int(other)
    
    
    def __lt__(self, other):
        other = self._format_other(other)
        return int(self) > int(other)
    
    def __le__(self, other):
        other = self._format_other(other)
        return int(self) >= int(other)
    
    def __int__(self):
        return [str(self) == level for level in self._HIARCHY].index(True)
    
    def __eq__(self, other):        
        return int(self) == int(self._format_other(other))
    
    def __str__(self):
        return self._plural_level_name
    
    def __repr__(self):
        return self.__str__()
    
    @property
    def _plural_level_name(self):
        if self.level in self._PLURAL_TO_SINGULAR.keys():
            return self.level
        else:
            return self._SINGULAR_TO_PLURAL[self.level]
        
    @property
    def singular(self):
        if self.level in self._SINGULAR_TO_PLURAL.keys():
            return self.level
        else:
            return self._PLURAL_TO_SINGULAR[self.level]
        
    # @property
    # def tags(self):
    #     factory = lambda tagnames: [OrthancDicomTag(name) for name in tagnames]
    #     if self == PATIENT_LEVEL:
    #         return factory(MAIN_PATIENT_TAGS)
    #     elif self == STUDY_LEVEL:
    #         return factory(MAIN_STUDY_TAGS)
    #     elif self==SERIES_LEVEL:
    #         return factory(MAIN_SERIES_TAGS)
    #     elif self==INSTANCE_LEVEL:
    #         return factory(MAIN_INSTANCE_TAGS)
    #     else:
    #         raise KeyError()
        
    @staticmethod
    def lowest_child_level(levels):
        return [level for level in levels if int(level) == max([int(li) for li in levels])][0]
    
    @staticmethod
    def highest_parent_level(levels):
        return [level for level in levels if int(level) == min([int(li) for li in levels])][0]

PATIENT_LEVEL = DicomLevel('Patient')
STUDY_LEVEL = DicomLevel('Study')
SERIES_LEVEL = DicomLevel('Series')
INSTANCE_LEVEL = DicomLevel('Instance')

# class TagStruct:    
#     def __init__(self, tagnames=None, level=None):
#         if level is None:
#             level = DicomLevel('Instance')
#         if tagnames is None:
#             tagnames = []
        
#         self.level = level
#         self.tags = [OrthancDicomTag(tagname) for tagname in tagnames]
#         super().__init__()
        
#     @property
#     def tagnames(self):
#         return [str(item) for item in self.tags]

#     def __dir__(self):
#         return super().__dir__() + self.tagnames
    
#     def __getattr__(self, attr):
#         if attr in self.tagnames:
#             return [item for item in self.tags if item.tagname==attr][0]
#         else:
#             raise AttributeError(attr)
    
#     def __iter__(self):
#         return self.tagnames.__iter__()
    
#     def __next__(self):
#         return self.tagnames.__next__()
        


@dataclass
class OrthancDicomTag():
    tagname: str
    PYDICOM_TAGNAMES = [value[4] for value in pydicom.datadict.DicomDictionary.values()]
    
    def __str__(self):
        return self.tagname
    
    def __repr__(self):
        return self.__str__()
    
    def __hash__(self):
        return self.pydicom_tag.__hash__()
    
    def __gt__(self, other):
        if isinstance(other, str):
            other = OrthancDicomTag(other)
        if isinstance(other, OrthancDicomTag):
            return self.tagname > other.tagname
        else:
            raise TypeError
            
    def __eq__(self, other):
        if isinstance(other, str):
            other = OrthancDicomTag(other)
        if isinstance(other, OrthancDicomTag):
            return self.tagname == other.tagname
        else:
            raise TypeError(type(other))
    
    def __lt__(self, other):
        if isinstance(other, str):
            other = OrthancDicomTag(other)
        if isinstance(other, OrthancDicomTag):
            return self.tagname < other.tagname
        else:
            raise TypeError
    
    @property
    def vr(self):
        return pydicom.datadict.dictionary_VR(self.pydicom_tag)
    
    @property
    def vm(self):
        return pydicom.datadict.dictionary_VM(self.pydicom_tag)
    
    @property
    def is_sequence(self):
        return self.vr == 'SQ'
    
    @property
    def orthanc_tag(self):
        return self.pydicom_to_orthanc_tag(self.pydicom_tag)
        
    @property
    def pydicom_tag(self):
        return pydicom.datadict.tag_for_keyword(self.tagname)
    
    @staticmethod
    def pydicom_to_orthanc_tag(tag):
        ptag = pydicom.tag.BaseTag(tag)
        grp = str(hex(ptag.group)).replace('0x', '')
        elem = str(hex(ptag.elem)).replace('0x', '')
        grp = '0'*(4-len(grp)) + grp
        elem = '0'*(4-len(elem)) + elem
        tag =  grp + '-' + elem
        return tag
    
    @classmethod
    def from_pydicom_tag(cls, tag):
        tagname = pydicom.datadict.keyword_for_tag(tag)
        return cls(tagname)
    
    @classmethod
    def from_orthanc_tag(cls, tag):
        tag = tag.replace('-', '')
        tag = tag.replace(',', '')
        tag = int(tag, 16)
        tagname = pydicom.datadict.keyword_for_tag(tag)
        return cls(tagname=tagname)
    

def _format_tags(json_dict):
    formatted = {}
    for key, value in json_dict.items():
        tag = OrthancDicomTag(key)
        if isinstance(value, list):
            formatted[tag] = [_format_tags(di) for di in value]
        else:
            formatted[tag] = value
    return formatted


def safe_parse_dicom_value(str_value, vr, vm='1'):
    if str_value is UNKNOWN:
        return None
    try:
        return parse_dicom_value(str_value, vr, vm)
    except:
        print(f'Could not parse {str_value} with VR: {vr} and VM: {vm}')
        return None

def parse_dicom_value(str_value, vr, vm='1', parse_date_time=False):
    vr = vr.upper()
    if vm != '1' and '\\' in str_value:
        return [parse_dicom_value(vi, vr, '1') for vi in str_value.split('\\')]
    
    if vr == 'AS':
        if not parse_date_time:
            return str_value
        # AS = 'nnnDWMY' D=days, W=weeks, M=months, Y=years
        # Calculate number of days
        
        dwmy = str_value[-1].upper()
        days_per_year = 365.2425
        if dwmy == 'D':
            mult = 1
        if dwmy == 'W':
            mult = 7
        if dwmy == 'M':
            mult = days_per_year / 12
        if dwmy == 'Y':
            mult = days_per_year
        
        return int(str_value[:-1]) * mult
    elif vr == 'DA':
        if not parse_date_time:
            return str_value
        # DA = 'YYYYMMDD'
        return parser.parse(str_value)
    
    elif vr == 'TM':
        if not parse_date_time:
            return str_value
        if '.' in str_value:
            str_value, frac_seconds = str_value.split('.')
        else:
            frac_seconds = '0'
        
        frac_seconds = float('0.' + frac_seconds)
        
        time = datetime.strptime(str_value, '%H%M%S')
        time += timedelta(seconds=frac_seconds)
        return time.time()

    elif vr == 'DT':
        if not parse_date_time:
            return str_value
        # DT = 'YYYYMMDDHHMMSS.FFFFFF&ZZXX'
        # ignore

        if '.' in str_value:
            dtstr, remainder = str_value.split('.')
            if '+' in remainder:
                frac_seconds, timezone = remainder.split('+')
                timezone = '+' + timezone
            elif '-' in remainder:
                frac_seconds, timezone = remainder.split('-')
                timezone = '-' + timezone
            else:
                frac_seconds = remainder
                timezone = None
        else:
            dtstr = str_value
            frac_seconds = '0'
            timezone = None
        
        frac_seconds = '0.' + frac_seconds

        if timezone is not None:
            hours       = timezone[1:3]
            minutes     = timezone[3:5]
            dtstr       += (timezone[0] + hours + ':' + minutes)
  
       
        dt = parser.parse(dtstr)
        dt += timedelta(seconds=float(frac_seconds))    
        return dt
    
    elif vr in ('AE', 'LT', 'PN', 'LT', 'SH', 'ST', 'UC', 'UI', 'UT', 'LO'):
        # actual strings
        return str_value.strip('\x00')
    elif vr in ('DS', 'FL', 'FD', 'OD', 'OF'):
        return float(str_value)
    elif vr in ('IS', 'OL', 'SL', 'SS', 'UL', 'US'):
        return int(str_value)
    elif vr in ('OB', 'OW', 'SQ', 'UN', 'UR', 'CS'):
        # Don't do anything with these VRs
        return str_value
        # raise TypeError(f'Cannot get a parsed value from VR: {vr}')
        
    else:
        raise TypeError(f'Unknown DICOM VR {vr}')



def dict_to_pydicom(orthanc_dict, ds=None, silent=True):
    if ds is None:
        ds = pydicom.Dataset()
    for key, value in orthanc_dict.items():
        key = OrthancDicomTag(key)
        tag = key.pydicom_tag
        
        
        if tag is None:
            if not silent:
                print(f'Could not parse dicom tag: {key}')
            continue
        
        vr = key.vr
        
        if key.is_sequence:
            nvalues = len(value)
            sq = pydicom.Sequence([Dataset() for i in range(0, nvalues)])
            ds[tag] = pydicom.DataElement(key.pydicom_tag, VR=vr, value=sq)
            for i in range(0, nvalues):
                dict_to_pydicom(value[i], ds=ds[tag][i], silent=silent)
        else:
            ds[tag] = pydicom.DataElement(key.pydicom_tag, vr, value)
    return ds
            
            


import logging

# json keys in orthanc response
TYPE =                          'Type'
ID =                            'ID'
PARENT  =                       'Parent'

UNKNOWN =                       '!UNKNOWN!'
LOG_LEVEL =                     logging.ERROR
PRIVATE =                       'private'
ASCENDING =                     'ascending'
DESCENDING =                    'descendig'

# when requesting more than this number of items a progressbar will be shown
SHOW_PROGRESS_BAR_COUNT =       1000





from setuptools import setup, find_packages
README = 'README.md'

DESCRIPTION = 'Pythonic interface to access DICOM data on an Orthanc Server'
NAME = 'simple-orthanc'

VERSION = "0.9.9.4"

from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()


setup(name=NAME,
      version=VERSION,
      description=DESCRIPTION,
      long_description=long_description,
      long_description_content_type='text/markdown',
      classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.6',
        'Topic :: Scientific/Engineering :: Medical Science Apps.',
        'Intended Audience :: Science/Research',
        'Natural Language :: English'
      ],
      keywords='image images orthanc simpleitk medical dicom',
      url='',
      author='M. Segbers',
      author_email='msegbers@gmail.com',
      license='MIT',
      packages=find_packages(),
      install_requires=[
          'pydicom',
          'pyyaml',
          'numpy',
          'requests',
          'python-dateutil',
          'tqdm',        
      ],
      extras_require={
          'SimpleITK': ['SimpleITK',
                        'sitktools @ git+https://gitlab.com/radiology/medical-physics/sitktools.git'
                        ],
          'Viewer': ['SimpleITK', 'simple-slice-viewer @ git+https://gitlab.com/radiology/medical-physics/simple-slice-viewer.git']
          },
         
      include_package_data=True,
      zip_safe=False)


# -*- coding: utf-8 -*-

#import constants


    
from simple_orthanc.constants import (ASCENDING, DESCENDING)



import simple_orthanc.logger
import simple_orthanc.constants
import simple_orthanc.dicom_tags
#from SimpleOrthanc.dicom_tags import MainDicomTags
from simple_orthanc.api import OrthancAPI
from simple_orthanc.interface import DicomLevels, OrthancInterface
from simple_orthanc.orthanc import Orthanc
from simple_orthanc.image import OrthancImage


import numpy as np
from simple_orthanc.dicom_tags import INSTANCE_LEVEL, SERIES_LEVEL, STUDY_LEVEL, PATIENT_LEVEL, OrthancDicomTag
from simple_orthanc.constants import  UNKNOWN, PARENT, ID
from simple_orthanc.helpers import show_progress
import json
from copy import copy
from dataclasses import dataclass, field
from typing import List
from itertools import chain
import io


@dataclass
class DicomLevels:
    Patients: List = field(default_factory=lambda: [])
    Studies: list = field(default_factory=lambda: [])
    Series: list = field(default_factory=lambda: [])
    Instances: list = field(default_factory=lambda: [])
    

    def __len__(self):
        return len(self.Patients) + len(self.Studies) + len(self.Series)\
            + len(self.Instances)
    
    def __iter__(self):
        return self.tolist().__iter__()
            
    def tolist(self):
        return self.Patients + self.Studies + self.Series + self.Instances
    
    def level(self, item):
        if item in self.Instances:
            return INSTANCE_LEVEL
        elif item in self.Series:
            return SERIES_LEVEL
        elif item in self.Studies:
            return STUDY_LEVEL
        elif item in self.Patients:
            return PATIENT_LEVEL
    def __str__(self):
        msg =\
            """
            Patients: {0}
            Studies: {1}
            Series: {2}
            Instances: {3}
            """
        return msg.format(self.Patients, self.Studies,
                          self.Series, self.Instances)
    
        
    
        
class OrthancInterface():
    """
    Class that keeps a cache of all orthanc oids and can interface a subset
    of oids (selection).
    """
    _index = None
    _selected = None
    _supports_numpy = None
    def __init__(self, api,  Patients=None, Studies=None,
                             Series=None, Instances=None):

        self.api = api
        self._generate_index()
        self.main_dicom_tags = self.get_main_dicom_tags()

    @property
    def supports_numpy(self):
        return self.api.version > '1.1.1'
    
    def get_main_dicom_tags(self):
        main_dicom_tags = DicomLevels()
        for level in (INSTANCE_LEVEL, SERIES_LEVEL, STUDY_LEVEL, PATIENT_LEVEL):
            tags = self.api.main_dicom_tags[level.singular].split(';')
            tags = [OrthancDicomTag.from_orthanc_tag(tag) for tag in tags]
            setattr(main_dicom_tags, str(level), tags)
        return main_dicom_tags
    
    def leveL_for_tag(self, tag):
        if tag in self.main_dicom_tags:
            return self.main_dicom_tags.level(tag)
        else:
            return INSTANCE_LEVEL
        
            
    def refresh(self):
        self._index = None
        self.reset()
    
    def reset(self):
        self._selected = None
        
    @property
    def index(self):
        # collection of all orthanc identifiers (oids) that are available
        # in Orthanc
        if self._index is None:
            self._generate_index()    
        return self._index
    
    def _generate_index(self):
        self._index = DicomLevels(**self.api.get_all_oids())
        
    @property
    def selected(self):
        # collection of orthanc identifiers that are selected by the user
        # property is set in the query method.
        if self._selected is None:
            # nothing selected all oids are in selection
            self._selected = copy(self.index)
        return self._selected

            
    def get_dicom_tag(self, tag, unique=True):
        # Return all values for a specific dicom tag for all selected 
        # Patients/Series/Studies/Instances on orthanc
        level = self.leveL_for_tag(tag)
        
        if isinstance(tag, str):
            tag = OrthancDicomTag(tag)
        
        if tag.is_sequence:
            values = self._get_dicom_sequence(tag, unique=unique)
            
        elif tag in self.main_dicom_tags and unique and level > INSTANCE_LEVEL:
            # limited set of tags that can be retrieved faster
            values = self._get_main_dicom_tag(tag)        
        else:
            # last resort get tag value for each instance
            values = self._get_instance_tag(tag)
            
        if unique:
            # Instances that do not have the tag will yield a UKNOWN value. When
            # unique is False keep the UKNOWN value to keep the 1-1 
            # correspondence between instances and values. If unique remove
            # UNKNOWN values.
            values = [value for value in values if value is not UNKNOWN]
            
        if not tag.is_sequence and unique:
            # preserve order when number of values won't change
            unique_values = list(set(values))
            if len(unique_values) < len(values):
                values = unique_values
        return values
    
    def _get_dicom_sequence(self, tag, unique=True):
        # Very Slow
        values = []

        for oid in show_progress(self.selected.Instances):
            all_tags = self.api.get_simplified_tags(oid)
            if str(tag) not in all_tags.keys():
                msg = 'Instance doesnt have Tag {0}'
                raise IndexError(msg.format(tag))
            values += [all_tags[str(tag)]]
        
        if unique:
            # convert sequences to json strings to select unique values
            values = list(set([json.dumps(value) for value in values]))
            values = [json.loads(value) for value in values]
        
        return values
    
    def _get_instance_tag(self, tag):
        values = []
        for oid in show_progress(self.selected.Instances):
            value = self.api.get_tag(instance_oid=oid, tag=str(tag))
            values += [value]
        return values
    
    def _get_main_dicom_tag(self, tag):
        level = self.leveL_for_tag(tag)
        oids = getattr(self.selected, str(level))
        
        values = []
        
        for oid in show_progress(oids):
            # requests are made for the specific dicom level. So
            # for PatientName there is no need to request the value of each
            # Instance
            request = self.api.request_oid(oid, level=str(level))
            response = self.api.get_json(request)
            if str(tag) in response[self.api.MAIN_DICOM_TAGS]:
                values += [response[self.api.MAIN_DICOM_TAGS][str(tag)]]
            else:
                values += [UNKNOWN]
        return values
    
    
    def get_header(self, instance_oid):
        request = self.api.request_simplified_tags(instance_oid)
        return self.api.get_json(request)
        
    def get_selected_headers(self):
        return [self.get_header(oid)\
                for oid in show_progress(self.selected.Instances)]
    
    def get_numpy(self, oid):
        if oid in self.index.Instances:
            level=str(INSTANCE_LEVEL)
        elif oid in self.index.Series:
            level=str(SERIES_LEVEL)
            
        request = self.api.request_numpy(oid, level=level)
        response = self.api.get(request)
        response.raise_for_status()
        image = np.load(io.BytesIO(response.content))
        return image
        
    
    def query(self, query=None, expand=True, select=True):
        # retrieve oids for a query and a specified level
        levels = [self.leveL_for_tag(tagname) for tagname in query.keys()]
        
        level = sorted(levels)[0]
        
    
        if level == PATIENT_LEVEL:
            level = STUDY_LEVEL
    
        data = {"Level": str(level), "Query": query, "Expand": expand}
    
        data = json.dumps(data)
        request = self.api.request_query()
        response = self.api.post_json(request, data)
        
        oids = self.decode_response(level, response)
        
        if select:
            self._selected = oids
        
        return oids
        
    
    def decode_response(self, level, response):
        oids = DicomLevels()
        if len(response) == 0:
            return oids
        if isinstance(response, list):
            if isinstance(response[0], dict):
                level_oids = [item[ID] for item in response]
            else:
                level_oids = response
                
            setattr(oids, str(level), level_oids)
    
        if level != PATIENT_LEVEL:
            key = level.parent_level.singular
            parent_key = PARENT + key
            
            parent_oids = [item[parent_key] for item in response]
            parent_oids = list(set(parent_oids))
            setattr(oids, str(level.parent_level), parent_oids)
    
        if level != INSTANCE_LEVEL:
            child_oids = []
            for item in response:
                coids = item[str(level.child_level)]
                if isinstance(coids, list):
                    child_oids += coids
                elif isinstance(coids, str):
                    child_oids += [coids]
            setattr(oids, str(level.child_level), child_oids)
            

        if level == INSTANCE_LEVEL:
            oids.Studies = self.get_parent_oids(oids.Series, SERIES_LEVEL)
    
        if level <= SERIES_LEVEL:
            oids.Patients = self.get_parent_oids(oids.Studies, STUDY_LEVEL)
            
        if level >= STUDY_LEVEL:
            oids.Instances = self.get_child_oids(oids.Series, SERIES_LEVEL)
        
        return oids
    
    def get_parent_oids(self, oid, level):
        if isinstance(oid, (list, tuple)):
            oids = [self.get_parent_oids(oii, level=level) for oii in oid]
            return list(set(oids))
        
        request = self.api.request_oid(oid=oid, level=level)
        data = self.api.get_json(request)
        parent_key = PARENT + str(level.parent_level.singular)
        return data[parent_key]
    
    def get_child_oids(self, oid, level):
        if isinstance(oid, (list, tuple)):
            oids = [self.get_child_oids(oii, level=level) for oii in oid]
            return list(set(chain(*oids)))
        
        request = self.api.request_oid(oid=oid, level=str(level))
        data = self.api.get_json(request)
        return data[level.child_level]
    
    


if __name__ == "__main__":
    pass



import json
from tqdm import tqdm
from simple_orthanc.constants import SHOW_PROGRESS_BAR_COUNT

def show_progress(iterable, count=SHOW_PROGRESS_BAR_COUNT):
    if len(iterable) <= count:
        return iterable
    else:
        return tqdm(iterable)

def deep_equal(item1, item2):
    return json.dumps(item1) == json.dumps(item2)
